#!/usr/bin/env python
# -*- coding: utf-8 -*-

#Generates numerical distribution functions that don't require the slow loading of scipy

from mpmath import *
import numpy as np

def tcdf(x, n):
  if type(x)==float or type(x)==np.float64:
    w = n/(x**2+n)
    I = betainc(0.5*n, 0.5,x2=w, regularized = True)
    dist = float(1-I)
    return dist
  if len(x)>100:
    raise RuntimeWarning(f"Dimension is {len(x)}. This function is not made for such large dimension. Consider using scipy. ")
  return np.array([tcdf(i, n) for i in x])


def norm(x,mu=0,s=1, cdf = True):
  if type(x)==float:
    if not cdf:
      return npdf(x, mu, s)
    f = 0.5*erfc((mu-x)/(s*2**0.5))
    return f
  if len(x)>100:
    raise RuntimeWarning(f"Dimension is {len(x)}. This function is not made for such large dimension. Consider using scipy. ")
  return np.array([norm(i, mu, s, cdf) for i in x])	

def chisq(x, k):
  if x<0:
    return 0
  c = gammainc(0.5*k, 0, 0.5*x)/gamma(0.5*k)
  return c

def fcdf(x, k1, k2):
  f = betainc(0.5*k1, 0.5*k2, x1 = 0, x2 = k1*x/(k1*x+k2), regularized = True)
  return f

def test_functions():
  from scipy import stats as scstats
  print(tcdf(3,5))
  print(scstats.t.cdf(3,5))

  print(norm(3,5,2))
  print(scstats.norm.cdf(3,5,2))

  print(chisq(3,5))
  print(scstats.chi2.cdf(3,5))

  print(fcdf(3,5,1))
  print(scstats.f.cdf(3,5,1))	

  print(norm(3,5,1, False))
  print(scstats.norm.pdf(3,5,1))		

#test_functions()
a=0


#!/usr/bin/env python
# -*- coding: utf-8 -*-

try:
  from . import slave
except:
  import slave

import os
import sys
from subprocess import Popen
import subprocess
import pickle
import datetime
from queue import Queue
from threading import Thread
import time
import tempfile
import random
import string
import hashlib
import psutil
import signal
import numpy as np
import __main__
import tempfile

#path = os.getcwd().replace('\\', '/') #uncomment to save to active folder
TMP_PATH = tempfile.gettempdir()

MIN_TIME = 1

class Parallel():
  """creates the slaves"""
  def __init__(self, max_nodes, run_parallel, callback_active):
    """module is a string with the name of the modulel where the
    functions you are going to run are """
    if max_nodes is None:
      self.cpu_count=os.cpu_count()
    else:
      self.cpu_count=max_nodes
    self.callback_active = callback_active
    n=self.cpu_count
    self.kill_warned = False
    self.is_parallel = run_parallel
    #directely in the simplest way, without any layers or multiprocessing (for debugging).
    self.kill_orphans()
    self.dict_file = create_temp_files(self.cpu_count)
    self.slaves=[Slave(run_parallel, n) for i in range(n)]
    self.final_results = {}
    self.n_tasks = {}
    self.t = time.time()
    self.pids = []
    self.master_pid = os.getpid()
    self.sum_running = {}
    for i in range(n):
      self.slaves[i].confirm(i) 
      pid=str(self.slaves[i].p_id)
      if int(i/5.0)==i/5.0:
        pid='\n'+pid
      self.pids.append(pid)
    pstr="""Multi core processing enabled using %s cores. \n
Master PID: %s \n
Slave PIDs: %s"""  %(n, self.master_pid,', '.join(self.pids))
    print (pstr)

  def send_dict(self, d):
    f = open(self.dict_file, 'wb')
    pickle.dump(d, f)  
    f.close()
    r = []		
    return self.async_send_receive('transfer dictionary', self.dict_file)


  def exec(self, tasks, name):
    #Handles at most as many tasks as cpus for the moment
    if tasks is None:
      tasks = [None]*self.cpu_count
    if type(tasks)==str:
      tasks = [tasks]*self.cpu_count
    self.n_tasks[name] = len(tasks)
    self.final_results[name] = [None]*self.n_tasks[name]
    self.sum_running[name] = len(tasks)
    return self.async_send_receive(name, tasks)

  def callback(self, name, outbox = {}, s_id = None, collect_finnished_procs = True):
    if not self.callback_active:
      return [{}]*self.cpu_count
    d=[]
    sumrun = -1
    while time.time() - self.t<MIN_TIME or sumrun<0:
      is_running = np.array(self.check_state(name))
      sumrun = sum(is_running)
      if sumrun<self.sum_running[name] :
        self.sum_running[name] = sumrun
        break #break if changed
      time.sleep(0.1)
    self.t = time.time()		


    if not s_id is None:
      self.slaves[s_id].send('callback',(name,outbox))
      return self.slaves[s_id].receive()	

    if not collect_finnished_procs:
      return self.async_send_receive('callback', (name, outbox))

    response = self.final_results[name]
    for i in range(self.n_tasks[name]):
      if (not is_running[i]) and (response[i] is None):
        r = self.collect(name, i)
        if not r is None and False:#set to true for debug
          print(f"collected sid:{i}, conv:{r['conv']}, f:{r['f']}, layer: {self.layer}")
        response[i] = r
    remaining_procs = np.arange(self.n_tasks[name])[np.array([i is None for i in response])]
    if len(remaining_procs)==0:
      return response
    response = list(response)
    r = self.async_send_receive('callback', (name, outbox), remaining_procs)
    for i in range(self.n_tasks[name]):
      if response[i] is None:
        response[i] = r[i]
    return response


  def collect(self, name, sid=None):
    if not sid is None:
      self.slaves[sid].send('collect', name)
      msg = self.slaves[sid].receive()	
      return 	msg

    return self.async_send_receive('collect', name)

  def async_send_receive(self,msg, obj, slave_list=[]):
    r = []
    q = Queue()

    if not type(obj) == list:
      obj = [obj]*self.cpu_count
    if len(slave_list)==0:
      slave_list = range(len(obj))		
    for i in slave_list:
      self.slaves[i].send(msg, obj[i])
      t = Thread(target = self.slaves[i].receive, args = (q, ), daemon= True)
      t.start()
    r = [{}]*len(obj)
    for i in slave_list:
      res, sid = q.get()
      r[sid] = res
    return r

  def count_alive(self):
    return sum(self.exec(None, 'check_state'))

  def check_state(self, name = None):
    return self.exec(name, 'check_state')



  def quit(self):
    print('quitting')
    for i in self.slaves:
      i.p.stdout.close()
      i.p.stderr.close()
      i.p.stdin.close()
      i.p.kill()
      i.p.cleanup()

  def kill_orphans(self):
    for proc in psutil.process_iter():
      try:
        self.kill_orphan(proc)
      except psutil.NoSuchProcess:
        pass


  def kill_orphan(self, proc):

    if proc.name() == 'python.exe' and False:#to shut down all python processes for debug, allmost never usefull.
      os.kill(proc.pid, signal.SIGTERM)

    if proc.name() == 'python.exe' and proc.cmdline()[-1]==(id_str()):
      os.kill(proc.pid, signal.SIGTERM)
      if not self.kill_warned:
        print("Due to the paralle processing, only one instance of a "
                                      "script can run at any time. If you want to run mulitple "
                                          "instances, create multiple scripts.")
        self.kill_warned = True
      print(f"killed pid {proc.pid}" )


def id_str():
  try:
    return __main__.__file__
  except:
    import IPython
    return str(id(IPython))

def makepath():
  fpath=os.path.join(TMP_PATH,'mp')
  os.makedirs(fpath, exist_ok=True)
  os.makedirs(fpath+'/slaves', exist_ok=True)	
  return fpath



class Slave():
  """Creates a slave"""
  def __init__(self, run_parallel, n):
    """Starts local worker"""
    self.n_nodes = n
    if run_parallel:		
      path = os.path.join(os.path.dirname(__file__),"init_node.py")
      command = [sys.executable, "-u",  path, id_str()]
      self.p = Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
      self.t = Transact(self.p.stdout,self.p.stdin)
      a=0

    else:
      qin, qout = Queue(), Queue()
      self.p = ThreadPopen(slave.run, (TransactThread(qout, qin),False))
      self.p.start()
      self.t = TransactThread(qin, qout)


  def confirm(self,slave_id):
    self.t.set(f'master {self.n_nodes}', self.n_nodes, self.t.fpath, slave_id)
    self.slave_id=slave_id
    self.p_id = self.receive()
    self.send('init_transact', (slave_id, self.t.fpath, self.n_nodes))

  def send(self,msg, obj):
    """Sends msg and obj to the slave"""
    if not self.p.poll() is None:
      raise RuntimeError(f'Subprocess {self.slave_id} has unexpectedly closed')
    self.t.send((msg, obj))     

  def receive(self, q = None):
    if not self.p.poll() is None:
      raise RuntimeError(f'Subprocess {self.slave_id} has unexpectedly closed')
    answ=self.t.receive()
    if q is None:
      return answ
    q.put((answ,self.slave_id))


  def kill(self):
    self.p.kill()

class Transact():
  """Local worker class"""
  def __init__(self,read, write):
    self.r = read
    self.w = write
    self.name = None
    self.slave_id = None
    self.time = time.time()
    self.f = None
    self.fpath = makepath()

  def set(self,name, n_nodes, path, slave_id):
    self.name = name
    self.n_nodes = n_nodes
    self.path = path
    self.slave_id = slave_id
    path = os.path.join(path,'debug')
    if not os.path.exists(path):
      os.mkdir(path)
    if False:#set ot True for debug
      path = os.path.join(path, f'{self.name}_{self.slave_id}_{n_nodes}.txt')
      self.f = open(path, 'w',1)		

  def send(self,msg):
    self.debug_output('send',msg)
    w=getattr(self.w,'buffer',self.w)
    pickle.dump(msg,w)
    w.flush()

  def receive(self):
    r=getattr(self.r,'buffer',self.r)
    u= pickle.Unpickler(r)
    try:
      msg = u.load()
      self.debug_output('recieve', msg)
      return msg
    except pickle.UnpicklingError as e:
      r.seek(0)
      u= pickle.Unpickler(r)
      return u.load()
    except EOFError as e:
      if e.args[0]=='Ran out of input':
        raise RuntimeError(
                          f"An error in {self.name} occured in sub-process {self.slave_id}.\n" 
                                f"Check the output in 'slave_errors.txt' in {self.fpath}\n"
                                f"You may try to run without multiprocessing\n {datetime.datetime.now()}"
                                )
      else:
        raise RuntimeError('EOFError:'+e.args[0])

  def debug_output(self, direction,msg):
    if self.f is None:
      return
    self.f.write(f'{direction}: {self.name}, {self.slave_id}\n{time.time()-self.time}:\n{str(msg)[:30]}\ntime:{time.time()}\n')	
    self.time = time.time()		


class ThreadPopen(Thread):
  """Starts local worker"""
  def __init__(self, target, args):
    super().__init__(target=target, args = args)
  def kill(self):
    raise RuntimeError("No kill procedure written yet")
  def poll(self):
    if self.is_alive():
      return None
    else:
      return 1

class TransactThread(Transact):
  """Thread version of Transact"""
  def __init__(self,read, write):
    self.r = read
    self.w = write
    self.name = None
    self.slave_id = None
    self.time = time.time()
    self.f = None	
    self.fpath = makepath()

  def set(self,name, n_nodes, path, slave_id):
    self.name = name
    self.n_nodes = n_nodes
    self.path = path
    self.slave_id = slave_id
    path = os.path.join(path,'debug')
    if not os.path.exists(path):
      os.mkdir(path)
    path = os.path.join(path, f'{self.name}_{self.slave_id}_{n_nodes}.txt')
    self.f = open(path, 'w',1)	

  def send(self,msg):
    self.debug_output('send',msg)
    self.w.put(msg) 

  def receive(self):
    msg = self.r.get()
    self.debug_output('recieve',msg)
    return msg

  def debug_output(self, direction,msg):
    return
    if self.f is None:
      return
    self.f.write(f'{direction}: {self.name}, {self.slave_id}\n{time.time()-self.time}:\n{str(msg)[:30]}\ntime:{time.time()}\n')	
    self.time = time.time()			


def write(f,txt):
  f = open(f, 'w')
  f.write(txt)
  f.flush()
  f.close()

def read(file):
  f = open(file, 'r')
  s = f.read()
  f.close()
  return s

def create_temp_files(cpu_count):
  number_of_files = 1

  f = get_file(TMP_PATH)
  try:#deleting old file
    folds = eval(read(f))
    for i in folds:
      os.remove(i)
  except Exception as e:
    print(e)
    a=0

  ftmp = []
  for i in range(number_of_files):
    f_name = gen_file_name()
    f_ = open(f_name,'w')
    f_.close()
    ftmp.append(f_name)
  write(f, str(ftmp))

  if number_of_files==1:
    ftmp = ftmp[0]
  return ftmp

def get_file(fpath):
  tdir = tempfile.gettempdir()
  f = os.path.join(tdir,'safdHLKk3223jd')
  #f = os.path.join(tdir,gen_file_name(fpath))
  return f

def gen_file_name(seed=False):
  #must use a seed that is constant also in an installed environment
  if not seed==False:
    f = os.path.join(os.sep.join(__file__.split(os.sep)[:-2]),os.path.join('.git','config'))
    n = len(seed)
    seed = int(hashlib.sha1(seed.encode('utf-8')).hexdigest(), 16)
    seed = os.path.getmtime(f) + seed
    random.seed(seed)
  else:
    random.seed()
  tdir = tempfile.gettempdir()
  fname = ''.join(random.choice(string.ascii_uppercase+string.ascii_lowercase + string.digits) for _ in range(20))
  fname = os.path.join(tdir,fname)
  return fname






#!/usr/bin/env python
# -*- coding: utf-8 -*-

#used for debugging


from . import functions as fu
from .likelihood import main as lgl
import numpy as np
import time
import os



def hess_debug(ll,panel,g,d):
  """Calculate the Hessian nummerically, using the analytical gradient. For debugging. Assumes correct and debugged gradient"""
  x=ll.args.args_v
  n=len(x)
  dx=np.identity(n)*d
  H=np.zeros((n,n))
  ll0=lgl.LL(x,panel)
  f0=g.get(ll0)
  for i in range(n):
    ll=lgl.LL(x+dx[i],panel)
    if not ll is None:
      f1=g.get(ll)
      H[i]=(f1-f0)/d


  return H

def grad_debug(ll,panel,d):
  """Calcualtes the gradient numerically. For debugging"""
  x=ll.args.args_v
  n=len(x)
  dx=np.abs(x.reshape(n,1))*d
  dx=dx+(dx==0)*d
  dx=np.identity(n)*dx

  g=np.zeros(n)
  f0=lgl.LL(x,panel)
  for i in range(n):
    for j in range(5):
      dxi=dx[i]*(0.5**j)
      f1=lgl.LL(x+dxi,panel)
      if not f1 is None:
        if not f1.LL is None:
          g[i]=(f1.LL-f0.LL)/dxi[i]
          break
  return g



def grad_debug_detail(f0,panel,d,llname,varname1,pos1=0):
  args1=lgl.copy_array_dict(f0.args.args_d)
  args1[varname1][pos1]+=d

  f0=lgl.LL(f0.args.args_d, panel)
  f1=lgl.LL(args1, panel)

  if type(llname)==list or type(llname)==tuple:
    ddL=(f1.__dict__[llname[0]].__dict__[llname[1]]-f0.__dict__[llname[0]].__dict__[llname[1]])/d
  else:
    ddL=(f1.__dict__[llname]-f0.__dict__[llname])/d
  return ddL

def test_c_armas(u_RE, var_ARMA, e_RE, panel, ll):
  N,T,k = e_RE.shape
  GAR_initvar =0
  if 'initvar' in ll.args.args_d:
    GAR_initvar = ll.GAR_1[0].reshape((1,T,1))*ll.args.args_d['initvar'][0]
  var_ARMA2 = GAR_initvar + panel.arma_dot.dot(ll.GAR_1MA,ll.h_val,ll)
  e_RE2 = panel.arma_dot.dot(ll.AMA_1AR,u_RE,ll)	
  print(f"Testsums arma: c:{np.sum(var_ARMA**2)}, py:{np.sum(var_ARMA2**2)}")
  print(f"Testsums e: c:{np.sum(e_RE**2)}, py:{np.sum(e_RE2**2)}")



def hess_debug_detail(f0,panel,d,llname,varname1,varname2,pos1=0,pos2=0):
  args1=lgl.copy_array_dict(f0.args.args_d)
  args2=lgl.copy_array_dict(f0.args.args_d)
  args3=lgl.copy_array_dict(f0.args.args_d)
  args1[varname1][pos1]+=d
  args2[varname2][pos2]+=d	
  args3[varname1][pos1]+=d
  args3[varname2][pos2]+=d
  f1=lgl.LL(args1, panel)
  f2=lgl.LL(args2, panel)
  f3=lgl.LL(args3, panel)
  if type(llname)==list:
    ddL=(f3.__dict__[llname[0]].__dict__[llname[1]]-f2.__dict__[llname[0]].__dict__[llname[1]]
                     -f1.__dict__[llname[0]].__dict__[llname[1]]+f0.__dict__[llname[0]].__dict__[llname[1]])/(d**2)
  else:
    ddL=(f3.__dict__[llname]-f2.__dict__[llname]-f1.__dict__[llname]+f0.__dict__[llname])/(d**2)
  return ddL



def LL_calc(self,panel):
  panel=self.panel
  X=panel.XIV
  matrices=set_garch_arch(panel,self.args.args_d)
  if matrices is None:
    return None		

  AMA_1,AMA_1AR,GAR_1,GAR_1MA=matrices
  (N,T,k)=X.shape
  #Idea for IV: calculate Z*u throughout. Mazimize total sum of LL. 
  u = panel.Y-fu.dot(X,self.args.args_d['beta'])
  e = fu.dot(AMA_1AR,u)
  e_RE = (e+self.re_obj_i.RE(e, panel)+self.re_obj_t.RE(e, panel))*panel.included[3]

  e_REsq =(e_RE**2+(e_RE==0)*1e-18) 
  grp = self.variance_RE(panel,e_REsq)#experimental

  W_omega = fu.dot(panel.W_a, self.args.args_d['omega'])

  lnv_ARMA = self.garch(panel, GAR_1MA, e_RE)

  lnv = W_omega+lnv_ARMA# 'N x T x k' * 'k x 1' -> 'N x T x 1'
  lnv+=grp
  self.dlnv_pos=(lnv<100)*(lnv>-100)
  lnv = np.maximum(np.minimum(lnv,100),-100)
  v = np.exp(lnv)*panel.a[3]
  v_inv = np.exp(-lnv)*panel.a[3]

  LL = self.LL_const-0.5*(lnv+(e_REsq)*v_inv)

  self.tobit(panel,LL)
  LL=np.sum(LL*panel.included[3])

  self.add_variables(matrices, u, e, lnv_ARMA, lnv, v, W_omega, grp,e_RE,e_REsq,v_inv)
  if abs(LL)>1e+100: 
    return None				
  return LL


def save_reg_data(ll, panel, fname = 'repr.csv', heading=True):
  #saves data neccessary to reproduce 
  N,T,k = panel.X.shape
  N,T,m = panel.W_a.shape
  heading = [f'X{i}' for i in range(k)]
  heading += [f'W{i}' for i in range(m)]
  heading += ['Y', 'e','u', 'var', 'LL','AMA_1AR', 'GAR_1MA' , 'coefs_names', 'coefs']
  a = np.concatenate((panel.X, panel.W_a, panel.Y, ll.e, ll.u, ll.var), 2)
  a = a.reshape((T, a.shape[2]))
  coefs = np.zeros((T,1))
  coef_names = np.array([['']]*T, dtype='<U20')
  coef_names[:]=''
  coefs[:len(ll.args.args_v),0] = ll.args.args_v
  coef_names[:len(ll.args.args_v),0] = ll.args.names_v
  a = np.concatenate((a, 
                            ll.LL_full[0].reshape((T,1)),
                                                ll.AMA_1AR[0].reshape((T,1))[::-1],
                                                ll.GAR_1MA[0].reshape((T,1))[::-1],
                                                coef_names,
                                                coefs
                                                ),1)
  a = np.concatenate(([heading], a),0)
  np.savetxt(fname, a, fmt='%s', delimiter=';')	


#!/usr/bin/env python
# -*- coding: utf-8 -*-

#This module contains classes used in the regression
from .. import likelihood as logl
from . import arguments
from .. import functions as fu

import numpy as np
import time



min_AC=0.000001

#todo: split up panel in sub classes
def posdef(a,da):
  return list(range(a,a+da)),a+da

class panel:
  def __init__(self,datainput,settings,pqdkm=None):

    self.input=datainput
    self.options=settings
    self.dataframe=datainput.dataframe
    if not pqdkm is None:
      self.pqdkm=pqdkm
    else:
      self.pqdkm=settings.pqdkm.value

  def init(self):
    self.initial_defs()
    self.arrayize()
    self.masking()
    self.lag_variables()
    self.final_defs()
    self.arma_dot=fu.ArmaDot()


  def initial_defs(self):
    if np.all(np.var(self.input.Y,0)==0):
      raise RuntimeError("No variation in Y")
    p,q,d,k,m=self.pqdkm
    self.orig_size=len(self.input.X)
    self.max_lags=self.input.max_lags
    self.lost_obs=max((p,q,self.max_lags))+max((m,k,self.max_lags))+d#+3
    self.nW,self.nZ,self.n_beta=len(self.input.W.columns),len(self.input.Z.columns),len(self.input.X.columns)
    self.define_h_func()
    self.Ld_inv=None
    if self.input.IDs is None:
      self.options.fixed_random_group_eff.value=0
    if self.input.IDs is None:
      self.options.fixed_random_group_eff.value=0		
    self.m_zero = False
    if  m==0 and k>0:
      self.m_zero = True
      k=0
      print("Warning: GARCH term removed since ARCH terms was set to 0")
    self.pqdkm=p,q,d,k,m

  def masking(self):

    #"initial observations" mask: 
    a=np.array([self.date_counter<self.T_arr[i] for i in range(self.N)])# sets observations that shall be zero to zero by multiplying it with the arrayized variable
    self.a=[None,None]
    self.a.extend([a.reshape(list(a.shape)[:-1]+[1]*i) for i in range(5)])
    #"after lost observations" masks: 
    self.T_i=np.sum(self.included[3],1).reshape((self.N,1,1))#number of observations for each i
    self.T_i=self.T_i+(self.T_i<=0)#ensures minimum of 1 observation in order to avoid division error. If there are no observations, averages will be zero in any case	
    self.N_t=np.sum(self.included[3],0).reshape((1,self.max_T,1))#number of observations for each t
    self.N_t=self.N_t+(self.N_t<=0)#ensures minimum of 1 observation in order to avoid division error. If there are no observations, averages will be zero in any case	
    self.group_var_wght=1-1/np.maximum(self.T_i-1,1)

  def final_defs(self):
    self.W_a=self.W*self.a[3]
    self.tot_lost_obs=self.lost_obs*self.N
    self.NT=np.sum(self.included[3])
    self.NT_before_loss=self.NT+self.tot_lost_obs
    if not hasattr(self,'n_dates'):
      self.number_of_RE_coef=0
      self.number_of_RE_coef_in_variance=0
      self.options.fixed_random_group_eff.set(0)
      self.options.fixed_random_time_eff.set(0)
      self.options.fixed_random_variance_eff.set(0)
    else:
      self.number_of_RE_coef=self.N*(self.options.fixed_random_group_eff.value>0)+self.n_dates*(self.options.fixed_random_time_eff.value>0)
      self.number_of_RE_coef_in_variance=(self.N*(self.options.fixed_random_group_eff.value>0)
                                                            +self.n_dates*(self.options.fixed_random_time_eff.value>0))*(self.options.fixed_random_variance_eff.value>0)
    self.args=arguments.arguments(self)
    self.df=self.NT-self.args.n_args-self.number_of_RE_coef-self.number_of_RE_coef_in_variance
    self.set_instrumentals()
    self.tobit()

  def lag_variables(self):
    T=self.max_T
    d=self.pqdkm[2]
    self.I=np.diag(np.ones(T))
    #differencing operator:
    if d==0:
      return
    L0=np.diag(np.ones(T-1),-1)
    Ld=(self.I-L0)
    Ld_inv=np.tril(np.ones((T,T)))
    for i in range(1,d):
      Ld=fu.dot(self.I-L0,Ld)
      Ld_inv=np.cumsum(Ld_inv,0)
    self.Ld_inv=Ld_inv
    #multiplication:
    self.X=self.lag_variable(self.X, Ld, d, True)
    self.Y=self.lag_variable(self.Y, Ld, d, False)
    if not self.Z is None:
      self.Z=self.lag_variable(self.Z, Ld, d, True)

  def lag_variable(self,X,Ld,d,recreate_intercept):
    X_out=fu.dot(Ld,X)*self.a[3]
    if self.input.has_intercept and recreate_intercept:
      X_out[:,:,0]=1
    X_out[:,:d]=0
    return X_out

  def params_ok(self,args):
    a=self.q_sel,self.p_sel,self.M_sel,self.K_sel
    for i in a:
      if len(i)>0:
        if np.any(np.abs(args[i])>0.999):
          return False
    return True

  def set_instrumentals(self):
    if self.input.Z.shape[1]==1:
      self.XIV=self.X
    else:
      self.XIV=self.X=self.Z
    return
    ll=logl.LL(self.args.args_init,self)
    ll.standardize(self)
    Z_st,Z_st_long=ll.standardize_variable(panel,self.Z)
    ZZ=fu.dot(Z_st,Z_st)
    ZZInv=np.linalg.inv(ZZ)
    ZX=fu.dot(Z_st,ll.X_st)
    ZZInv_ZX=fu.dot(ZZInv, ZX)
    self.XIV=fu.dot(self.Z, ZZInv_ZX)#using non-normalized first, since XIV should be unnormalized.	

  def subtract_means(self,X,Y,Z):
    subtract=self.options.subtract_means.value
    if not subtract:
      return X,Y,Z
    X=X-np.mean(X,0)
    if self.input.has_intercept:
      X[:,0]=1
    Y=Y-np.mean(Y,0)
    if self.input.Z.shape[1]==1:
      return X,Y,Z
    Z=Z-np.mean(Z,0)
    Z[:,0]=1		

  def is_single(self):
    IDs,t=self.input.IDs_num,self.input.timevar_num
    try:
      if (np.all(IDs.iloc[0]==IDs) or np.all(t.iloc[0]==t)):
        return True
    except:
      return True
    return False


  def arrayize(self):
    """Splits X and Y into an arry of equally sized matrixes rows equal to the largest for each IDs"""

    X, Y, W, IDs,timevar,Z=[to_numpy(i) for i in 
                                        (self.input.X, self.input.Y, self.input.W, self.input.IDs_num,self.input.timevar_num,self.input.Z)]
    X,Y,Z=self.subtract_means(X,Y,Z)
    NT,k=X.shape
    self.total_obs=NT
    if self.is_single():
      if not np.all(timevar[:,0]==np.sort(timevar[:,0])):#remove in production
        raise RuntimeError("The time variable is not sorted!?!")			
      self.X=np.array(X.reshape((1,NT,k)))
      self.Y=np.array(Y.reshape((1,NT,1)))
      NT,kW=W.shape
      self.W=np.array(W.reshape((1,NT,kW)))
      NT,kZ=Z.shape
      self.Z=None
      if Z.shape[1]>1:#instrumental variables used
        self.Z=np.array(Z.reshape((1,NT,kZ)))
      self.time_map=None
      self.map=np.arange(NT).reshape(1,NT)
      self.N=1
      self.max_T=NT
      self.T_arr=np.array([[NT]])
      self.date_counter=np.arange(self.max_T).reshape((self.max_T,1))
      included=np.array([(self.date_counter>=self.lost_obs)*(self.date_counter<self.T_arr[i]) for i in range(self.N)])
      self.date_count_mtrx=None
      self.date_count=None
      self.idincl=np.array([True])
      if not np.all(timevar[:,0]==np.sort(timevar[:,0])):#remove in production
        raise RuntimeError("The arrayize procedure has unsorted the time variable!?!")					
    else:
      sel,ix=np.unique(IDs,return_index=True)
      N=len(sel)
      sel=(IDs.T==sel.reshape((N,1)))
      T=np.sum(sel,1)
      self.max_T=np.max(T)
      self.idincl=T>self.lost_obs+self.options.min_group_df.value
      self.X=arrayize(X, N,self.max_T,T, self.idincl,sel)
      self.Y=arrayize(Y, N,self.max_T,T, self.idincl,sel)
      self.W=arrayize(W, N,self.max_T,T, self.idincl,sel)
      self.Z=arrayize(Z, N,self.max_T,T, self.idincl,sel)		
      self.N=np.sum(self.idincl)
      varmap=np.arange(NT).reshape(NT,1)
      self.map=arrayize(varmap, N,self.max_T,T, self.idincl,sel,dtype=int).reshape((self.N,self.max_T))
      self.T_arr=T[self.idincl].reshape((self.N,1))
      self.date_counter=np.arange(self.max_T).reshape((self.max_T,1))
      included=np.array([(self.date_counter>=self.lost_obs)*(self.date_counter<self.T_arr[i]) for i in range(self.N)])
      if len(included)<5:
        raise RuntimeError(f"{len(included)} valid observations, cannot perform panel analysis. Try without panel (don't specify ID and time)")
      self.get_time_map(timevar, self.N,T, self.idincl,sel,included)

      if np.sum(self.idincl)<len(self.idincl):
        idname=self.input.IDs.columns[0]
        idremoved=np.array(self.input.IDs)[ix,0][self.idincl==False]
        s = formatarray(idremoved,90,', ')
        print(f"The following {idname}s were removed because of insufficient observations:\n %s" %(s))
      #remove in production. Checking sorting:
      tvar=arrayize(timevar, N,self.max_T,T, self.idincl,sel)
      a=[tvar[i][0][0] for i in self.date_map]
      if not np.all(a==np.sort(a)):	
        raise RuntimeError("It seems the data is not properly sorted on time")

    zeros=np.zeros((self.N,self.max_T,1))
    ones=np.ones((self.N,self.max_T,1))			
    self.included=[None,None]
    self.zeros=[None,None]
    self.ones=[None,None]
    self.included.extend([included.reshape(list(included.shape)[:-1]+[1]*i) for i in range(5)])		
    self.zeros.extend([zeros.reshape(list(zeros.shape)[:-1]+[1]*i) for i in range(5)])	
    self.ones.extend([ones.reshape(list(ones.shape)[:-1]+[1]*i) for i in range(5)])	


  def tobit(self):
    """Sets the tobit threshold"""
    tobit_limits=self.options.tobit_limits.value
    if tobit_limits is None:
      return
    if len(tobit_limits)!=2:
      print("Warning: The tobit_limits argument must have lenght 2, and be on the form [floor, ceiling]. None is used to indicate no active limit")
    if (not (tobit_limits[0] is None)) and (not( tobit_limits[1] is None)):
      if tobit_limits[0]>tobit_limits[1]:
        raise RuntimeError("floor>ceiling. The tobit_limits argument must have lenght 2, and be on the form [floor, ceiling]. None is used to indicate no active limit")
    g=[1,-1]
    self.tobit_I=[None,None]
    self.tobit_active=[False, False]#lower/upper threshold
    desc=['tobit_low','tobit_high']
    for i in [0,1]:
      self.tobit_active[i]=not (tobit_limits[i] is None)
      if self.tobit_active[i]:
        if np.sum((g[i]*self.Y<g[i]*tobit_limits[i])*self.a[3]):
          print("Warning: there are observations of self.Y outside the non-censored interval. These will be ignored.")
        I=(g[i]*self.Y<=g[i]*tobit_limits[i])*self.a[3]
        self.Y[I]=tobit_limits[i]
        if np.var(self.Y)==0:
          raise RuntimeError("Your tobit limits are too restrictive. All observationss would have been cencored. Cannot run regression with these limits.")
        self.tobit_I[i]=I
        if np.sum(I)>0 and np.sum(I)<self.NT and False:#avoiding singularity #Not sure why this is here, shuld be deleted?
          self.X=np.concatenate((self.X,I),2)
          self.input.X_names.append(desc[i])

  def get_time_map(self,timevar, N,T_count, idincl,sel,incl):
    if timevar is None:
      return None
    N,T,k=incl.shape
    unq,ix=np.unique(timevar,return_inverse=True)
    t=arrayize(np.array(ix).reshape((len(timevar),1)), 
                           N,self.max_T,T_count, idincl,sel,int)#maps N,T -> unique date
    grp_cnt=incl*np.arange(N).reshape(N,1,1)
    t_cnt=incl*np.arange(T).reshape(1,T,1)
    incl=incl[:,:,0]
    t=np.concatenate((t[incl],  grp_cnt[incl], t_cnt[incl]),1)
    a=np.argsort(t[:,0])
    t=t[a]#three columns: unique date index, group number, day sequence


    tid=t[:,0]#unique date
    t_map=[[] for i in range(np.max(tid)+1)]#all possible unique dates
    for i in range(len(tid)):
      t_map[tid[i]].append(t[i,1:])#appends group and day sequence
    t_map_tuple=[]
    tcnt=[]
    self.date_count_mtrx=np.zeros((N,T,1))
    for i in range(len(t_map)):
      a=np.array(t_map[i]).T#group and day sequence for unique date i
      if len(a):
        m=(tuple(a[0]),tuple(a[1]))#group and day sequence reference tuple
        n_t=len(a[0])#number of groups at this unique date
        t_map_tuple.append(m)	#group and day reference tuple for the data matrix, for each unique date
        tcnt.append(n_t) #count of groups at each date
        self.date_count_mtrx[m]=n_t#timeseries matrix of the group count


    #A full random effects calculation is infeasible because of complexity and computing costs. 
    #A quazi random effects weighting is used. It  is more conservative than the full
    #RE weight theta=1-sd_pooled/(sd_pooled+sd_within/T)**0.5
    #If the weights are too generous, the RE adjustment may add in stead of reducing noise. 
    n=len(tcnt)
    self.n_dates=n
    self.date_count=np.array(tcnt).reshape(n,1,1)
    self.date_map=t_map_tuple


  def get_time_map2(self,timevar, N,T_count, idincl,sel):
    if timevar is None:
      return None
    unq,ix=np.unique(timevar,return_inverse=True)
    n_dates=len(unq)
    t=arrayize(np.array(ix).reshape((n_dates,1)), N,self.max_T,T_count, idincl,sel,int)#maps N,T -> unique date
    N,T,k=t.shape


    t_map=[[] for i in range(n_dates)]#all possible unique dates
    for i in range(len(tid)):
      t_map[tid[i]].append(t[i,1:])#appends group and day sequence
    t_map_tuple=[]
    tcnt=[]
    self.date_count_mtrx=np.zeros((N,T,1))
    for i in range(len(t_map)):
      a=np.array(t_map[i]).T#group and day sequence for unique date i
      if len(a):
        m=(tuple(a[0]),tuple(a[1]))#group and day sequence reference tuple
        n_t=len(a[0])#number of groups at this unique date
        t_map_tuple.append(m)	#group and day sequence reference tuple, for each unique date
        tcnt.append(n_t) #count of groups at each date
        self.date_count_mtrx[m]=n_t#timeseries matrix of the group count


    #A full random effects calculation is infeasible because of complexity and computing costs. 
    #A quazi random effects weighting is used. It  is more conservative than the full
    #RE weight theta=1-sd_pooled/(sd_pooled+sd_within/T)**0.5
    #If the weights are too generous, the RE adjustment may add in stead of reducing noise. 
    n=len(tcnt)
    self.n_dates=n
    self.date_count=np.array(tcnt).reshape(n,1,1)
    self.date_map=t_map_tuple


  def define_h_func(self):
    h_def="""
def h(e,z):
	e2			=	e**2+1e-5
	h_val		=	np.log(e2)	
	h_e_val		=	2*e/e2
	h_2e_val	=	2/e2-4*e**2/e2**2

	return h_val,h_e_val,h_2e_val,None,None,None
		"""	
    h_definition=self.options.h_function.value
    if h_definition is None:
      h_definition=h_def
    d=dict()
    try:
      exec(h_definition,globals(),d)
      ret=d['h'](1,1)
      if len(ret)!=6:
        raise RuntimeError("""Your custom h-function must return exactly six arguments
				(h, dh/dx and ddh/dxdx for both e and z. the z return values can be set to None)""")
      self.h_def=h_definition
    except Exception as e:
      print('Something is wrong with your custom function, default is used:'+ str(e))
      exec(h_def,globals(),d)
      self.h_def=h_def

    self.z_active=True
    for i in ret[3:]:
      self.z_active=self.z_active and not (i is None)	


  def mean(self,X,axis=None):
    dims=list(X.shape)
    dims[2:]=[1]*(len(dims)-2)
    #X=X*self.included.reshape(dims)
    if axis is None:
      return np.sum(X)/self.NT
    if axis==1:
      dims.pop(1)
      return np.sum(X,1)/self.T_i.reshape(dims)
    if axis==0:
      dims.pop(0)
      return np.sum(X,0)/self.N_t.reshape(dims)
    if axis==(0,1):
      return np.sum(np.sum(X,0),0)/self.NT

  def var(self,X,axis=None,k=1,mean=None,included=None):
    dims=list(X.shape)
    dims_m=np.array(X.shape)
    dims[2:]=[1]*(len(dims)-2)	
    if included is None:
      a=self.included[len(dims)]
    else:
      a=included
    if mean is None:
      m=self.mean(X*a, axis)
    else:
      m=mean

    if axis==None:
      Xm=(X-m)*a
      return np.sum(Xm**2)/(self.NT-k)

    if axis==1:
      dims_m[1]=1
      m=m.reshape(dims_m)
      Xm=(X-m)*a
      dims.pop(1)
      return np.sum((Xm)**2,1)/np.maximum(self.T_i-k,1).reshape(dims)
    if axis==0:
      dims_m[0]=1		
      m=m.reshape(dims_m)
      Xm=(X-m)*a
      dims.pop(0)
      return np.sum((Xm)**2,0)/np.maximum(self.N_t-k,1).reshape(dims)
    if axis==(0,1):
      dims_m[0:2]=1
      m=m.reshape(dims_m)
      Xm=(X-m)*a			
      return np.sum((Xm)**2,axis)/(self.NT-k)

def arrayize(X,N,max_T,T,idincl,sel,dtype=None):
  if X is None:
    return None
  NT,k=X.shape
  if dtype is None:
    Xarr=np.zeros((N,max_T,k))
  else:
    Xarr=np.zeros((N,max_T,k),dtype=dtype)
  T_used=[]
  k=0
  for i in range(len(sel)):
    if idincl[i]:
      Xarr[k,:T[i]]=X[sel[i]]
      k+=1
  Xarr=Xarr[:k]
  return Xarr






def to_numpy(x):
  x=np.array(x)
  if len(x.shape)==2:
    return x
  return x.reshape((len(x),1))

def formatarray(array,linelen,sep):
  s=sep.join([str(i) for i in array])
  s='\n'.join(s[n:n + linelen] for n in range(0, len(s), linelen))	
  return s



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#This module handle callbacks



class CallBack:
  def __init__(self, inbox, outbox):
    self.inbox = inbox
    self.outbox = outbox
    self.outbox['quit'] = False


  def callback(self, **keywordargs):
    #print(f'writing to {id(self.outbox)}')
    for k in keywordargs:
      self.outbox[k] = keywordargs[k]
    return dict(self.inbox)















#!/usr/bin/env python
# -*- coding: utf-8 -*-

#This module handle interfacing for various output paltforms


try:
  get_ipython()#fails if not in IPython environment
  import IPython
except NameError as e:
  IPython = None

from pydoc import importfile
import os
path = os.path.dirname(__file__)
from . import output

import os
import numpy as np
import time



WEB_PAGE='paneltime.html'
TMP_PAGE='tmphtml'
pic_num=[1]





def get_channel(window,exe_tab,panel, console_output):
  if console_output:
    return console(panel)
  if not window is None:#tkinter gui
    return tk_widget(window,exe_tab,panel)
  if not IPython is None:
    n=IPython.get_ipython().__class__.__name__
    if n=='ZMQInteractiveShell':
      return web_output(True,panel)
  try:
    return web_output(False,panel)
  except:
    pass
  return console(panel)

class web_output:
  def __init__(self,Jupyter,panel):
    global webbrowser
    import webbrowser
    self.panel=panel	
    self.Jupyter=Jupyter
    if not Jupyter:
      self.save_html(get_web_page('None', 'None', 'None', '', True))
      if panel.options.web_open_tab.value:
        webbrowser.open(WEB_PAGE, new = 2)
    self.output_set = False
    #from . import charts
    #self.charts = charts.ProcessCharts(panel)

  def set_progress(self,perc, text, task):
    return True

  def set_output_obj(self,ll, comput, dx_norm):
    "sets the outputobject in the output" 
    if self.output_set:
      return		
    self.output=output.output(ll,self.panel, comput, dx_norm)
    self.output_set = True


  def update(self,comput, its, ll, incr, dx_norm):
    if self.panel.options.supress_output.value:
      return
    self.output.update(comput,its, ll, incr, dx_norm)
    self.its=its
    self.reg_table=self.output.reg_table()
    tbl,llength=self.reg_table.table(4,'(','HTML',True,
                                                 show_direction=True,
                                                           show_constraints=True)		
    web_page=get_web_page(ll.LL, 
                                      ll.args.args_v, 
                                                          dx_norm,
                                                          tbl,
                                                          self.Jupyter==False)
    #self.charts.save_all(ll)
    if self.Jupyter:
      IPython.display.clear_output(wait=True)
      display(IPython.display.HTML(web_page))
    else:
      self.save_html(web_page)


  def save_html(self,htm_str):
    self.f = open(WEB_PAGE, "w")
    self.f.write(htm_str)
    self.f.close()


  def print_final(self, msg, fret, conv, t0, xsol, its, node):
    print(msg)
    print(f"LL={fret}  success={conv}  t={time.time()-t0}  its: {its}   node: {node}")
    print(xsol)	

class console:
  def __init__(self,panel):
    self.panel=panel
    self.output_set = False

  def set_progress(self,perc,text, task):
    if task=='done':
      print(text)
    #perc = f'{int(perc*100)}%'.ljust(5)
    #print(f"{perc} - {task}: {text}")
    return True

  def set_output_obj(self,ll, comput, dx_norm):
    if self.output_set:
      return
    self.output=output.output(ll,self.panel, comput, dx_norm)
    self.output_set = True
    self.comput = comput

  def update(self,comput, its,ll,incr, dx_norm):
    if self.panel.options.supress_output.value:
      return
    self.ll = ll
    self.incr = incr
    self.dx_norm = dx_norm


  def print_final(self, msg, fret, conv, t0, xsol, its, node):
    if self.panel.options.supress_output.value:
      return
    print(msg)
    if self.output_set:
      self.output.update(self.comput,its, self.ll, 0, self.dx_norm)
      self.its=its
      self.reg_table=self.output.reg_table()	
      tbl,llength=self.reg_table.table(5,'(','CONSOLE',False,
                                                         show_direction=False,
                                                                   show_constraints=False)	
      print(tbl)
    else:
      print(f"LL={fret}  success={conv}  t={time.time()-t0}  its: {its}   node: {node}")
      print(xsol)		

class tk_widget:
  def __init__(self,window,exe_tab,panel):
    self.panel=panel
    self.tab=window.main_tabs._tabs.add_output(exe_tab)
    self.set_progress=self.tab.progress_bar.set_progress
    self.output_set = False

  def set_output_obj(self,ll, comput, dx_norm):
    if self.output_set:
      return		
    self.tab.set_output_obj(ll,self.panel, comput, dx_norm)
    self.output = self.tab.output
    self.output_set = True

  def update(self,comput, its, ll, incr, dx_norm):
    if self.panel.options.supress_output.value:
      return
    self.tab.update(self.panel, comput,its, ll, incr, dx_norm)

  def print_final(self, msg, fret, conv, t0, xsol, its, node):
    if self.panel.options.supress_output.value:
      return
    print(msg)
    print(f"LL={fret}  success={conv}  t={time.time()-t0} its: {its}")
    print(xsol)	

def get_web_page(LL, args, comput,tbl,auto_update):
  au_str=''
  if auto_update:
    au_str="""<meta http-equiv="refresh" content="1" >"""
  img_str=''
  pic_num[0]+=1
  if os.path.isfile('img/chart0.png'):
    img_str=(f"""<img src="img/histogram.png"><br>\n"""
                         f"""<img src="img/correlogram.png"   ><br>\n"""
                                f"""<img src="img/correlogram_variance.png"   >""")
  return f"""
<meta charset="UTF-8">
{au_str}
<head>
<title>paneltime output</title>
</head>
<style>
p {{
               margin-left: 60px;
               max-width: 980px;
               font-family: "Serif";
               text-align: left;
               color:#063f5c;
               font-size: 16;
}}
h1 {{
               margin-left: 20px;
               max-width: 980px;
               font-family: "Serif";
               text-align: left;
               color:black;
               font-size: 25;
               font-weight: bold;
}}

table.head {{
               font-family: "Serif";
               text-align: right;
               color:black;
               padding-left: 0px;
               font-size: 16;
}}
td.h:nth-child(odd) {{
               background: #CCC
}}
td {{
               padding-left: 0px;
}}
table {{
               font-family: "Serif";
               text-align: right;
               color:black;
               font-size: 16;
}}

th {{
border-collapse: collapse;
               border-bottom: double 3px;
               padding-left: 0px;
               white-space: nowrap;
}}
</style>
<body>
<div style='position:absolute;float:right;top:0;right:0'>
{img_str}
</div>
{tbl}
</body>
</html> """	




#!/usr/bin/env python
# -*- coding: utf-8 -*-


from . import calculus
from . import calculus_functions
from . import function
from . import main


LL = main.LL
gradient = calculus.gradient
hessian = calculus.hessian
func_gradent = function.gradient
func_hessian = function.hessian
set_garch_arch = main.set_garch_arch




#!/usr/bin/env python
# -*- coding: utf-8 -*-

#This module contains the argument class for the panel object
from ..output import stat_functions as stat
from .. import likelihood as logl
from .. import random_effects as re
from .. import functions as fu

import numpy as np





class arguments:
  """Sets initial arguments and stores static properties of the arguments"""
  def __init__(self,panel):
    p, q, d, k, m=panel.pqdkm
    self.initial_user_defined = False
    self.categories=['beta','rho','lambda','gamma','psi','omega']
    if panel.options.include_initvar.value:
      self.categories+=['initvar']
    if panel.z_active:
      self.categories+=['z']
    self.mu_removed=True
    if not self.mu_removed:
      self.categories+=['mu']
    self.make_namevector(panel,p, q, k, m)
    initargs=self.initargs(p, d, q, m, k, panel)
    self.position_defs(initargs)
    self.set_init_args(panel,initargs)
    self.get_user_constraints(panel)


  def get_user_constraints(self,panel):
    e="User contraints must be a dict of dicts or a string evaluating to that, on the form of ll.args.dict_string. User constraints not applied"
    if type(panel.options.user_constraints.value)==dict:
      self.user_constraints=panel.options.user_constraints.value
    else:
      if panel.options.user_constraints.value is None or panel.options.user_constraints.value=='':
        self.user_constraints={}
        return
      try:
        self.user_constraints=eval(panel.options.user_constraints.value)
      except SyntaxError:
        print(f"Syntax error: {e}")
        self.user_constraints={}
        return			
      except:
        print(e)
        self.user_constraints={}
        return
    if not panel.z_active and 'z' in self.user_constraints:
      self.user_constraints.pop('z')	
    if panel.options.include_initvar.value  and 'initvar' in self.user_constraints:
      self.user_constraints.pop('initvar')	
    for grp in self.user_constraints:
      try:
        self.user_constraints[grp] = np.array(self.user_constraints[grp]).flatten()
        test = 1.0 * self.user_constraints[grp]
      except:
        print(f'Failed to apply user constraints for {grp}')
        self.user_constraints.pop('grp')

    for grp in self.user_constraints:
      if not grp in self.caption_d:
        print(f"Constraint on {grp} not applied, {grp} not in arguments")
      if len(self.caption_d[grp]) < len(self.user_constraints[grp]):
        print(f"Not all constraints on {grp} applied, more constraints than arguments for {grp}")



  def initargs(self,p,d,q,m,k,panel):

    args=dict()
    args['beta']=np.zeros((panel.X.shape[2],1))
    args['omega']=np.zeros((panel.W.shape[2],1))
    args['rho']=np.zeros(p)
    args['lambda']=np.zeros(q)
    args['psi']=np.zeros(m)
    args['gamma']=np.zeros(k)
    args['omega'][0][0]=0
    args['mu']=np.array([])
    if panel.options.include_initvar.value:
      args['initvar']=np.zeros(1)
    args['z']=np.array([])			


    if m>0 and panel.z_active:
      args['z']=np.array([1e-09])	

    if panel.N>1 and not self.mu_removed:
      args['mu']=np.array([0.0001])			


    return args

  def set_init_args(self,panel,initargs=None, default = True):
    p, q, d, k, m=panel.pqdkm

    if initargs is None:
      initargs = self.initargs(p, d, q, m, k, panel)

    #de2=np.roll(e**2,1)-e**2
    #c=stat.correl(np.concatenate((np.roll(de2,1),de2),2),panel)[0,1]
    beta,omega = self.set_init_regression(initargs,panel, default)
    self.args_start=self.create_args(initargs,panel)
    self.args_init=self.create_args(initargs,panel)
    self.set_restricted_args(p, d, q, m, k,panel,omega,beta)
    self.n_args=len(self.args_init.args_v)


  def set_restricted_args(self,p, d, q, m, k, panel,omega,beta):
    args_restricted=self.initargs(p, d, q, m, k, panel)
    args_OLS=self.initargs(p, d, q, m, k, panel)	
    
    args_restricted['beta'][0][0]=np.mean(panel.Y)
    args_OLS['beta']=beta
    
    v = panel.var(panel.Y)
    if panel.options.EGARCH.value==0:
      args_restricted['omega'][0][0]= v
      args_OLS['omega'][0][0]=omega
    else:
      args_restricted['omega'][0][0]=np.log(v)
      args_OLS['omega'][0][0]=np.log(omega)
      
    self.args_restricted=self.create_args(args_restricted,panel)
    self.args_OLS=self.create_args(args_OLS,panel)


  def create_null_ll(self,panel):
    if not hasattr(self,'LL_OLS'):
      self.LL_OLS=logl.LL(self.args_OLS,panel).LL
      self.LL_null=logl.LL(self.args_restricted,panel).LL	

  def position_defs(self,initargs):
    """Defines positions in vector argument"""

    self.positions=dict()
    self.positions_map=dict()#a dictionary of indicies containing the string name and sub-position of index within the category
    k=0
    for i in self.categories:
      n=len(initargs[i])
      rng=range(k,k+n)
      self.positions[i]=rng
      for j in rng:
        self.positions_map[j]=[0,i,j-k]#equation,category,relative position
      k+=n


  def conv_to_dict(self,args):
    """Converts a vector argument args to a dictionary argument. If args is a dict, it is returned unchanged"""
    if type(args)==dict:
      return args
    if type(args)==list:
      args=np.array(args)			
    d=dict()
    k=0
    for i in self.categories:
      n=len(self.positions[i])
      rng=range(k,k+n)
      d[i]=np.array(args[rng])
      if i=='beta' or i=='omega':
        d[i]=d[i].reshape((n,1))
      k+=n
    return d


  def conv_to_vector(self,args):
    """Converts a dict argument args to vector argument. if args is a vector, it is returned unchanged.\n
    If args=None, the vector of self.args_init is returned"""
    if type(args)==list or type(args)==np.ndarray:
      return np.array(args)
    v=np.array([])
    for i in self.categories:
      s=np.array(args[i])
      if len(s.shape)==2:
        s=s.flatten()
      if len(s)>0:
        v=np.concatenate((v,s))
    return v


  def make_namevector(self,panel,p, q, k, m):
    """Creates a vector of the names of all regression varaibles, 
    including variables, ARIMA and GARCH terms. This defines the positions
    of the variables througout the estimation."""
    d, names_d = {}, {}
    captions=list(panel.input.X.keys())#copy variable names
    d['beta']=list(captions)
    c=[list(captions)]
    names = [f'x{i}' for i in range(panel.n_beta)]
    names_d['beta'] = list(names)
    add_names(p,'rho%s    AR    p','rho',d,c,captions, names, names_d)
    add_names(q,'lambda%s MA    q','lambda',d,c,captions, names, names_d)
    add_names(k,'gamma%s  GARCH k','gamma',d,c,captions, names, names_d)
    add_names(m,'psi%s    ARCH  m','psi',d,c,captions, names, names_d)


    d['omega']=list(panel.input.W.keys())
    captions.extend(panel.input.W.keys())

    names_d['omega'] = [f'omega{i}' for i in range(panel.nW)]
    names.extend(names_d['omega'])

    c.append(d['omega'])
    if panel.options.include_initvar.value:
      d['initvar'] = ['Initial variance']
      captions.extend(d['initvar'])
      names_d['initvar'] = d['initvar']
      names.extend(names_d['initvar'])
      c.append(d['initvar'])
    
    if m>0:
      if panel.N>1 and not self.mu_removed:
        d['mu']=['mu (var.ID eff.)']
        captions.extend(d['mu'])
        names_d['mu']=['mu']
        names.extend(d['mu'])				
        c.append(d['mu'])
      if panel.z_active:
        d['z']=['z in h(e,z)']
        captions.extend(d['z'])
        names_d['z']=['z']
        names.extend(d['z'])					
        c.append(d['z'])

    self.caption_v=captions
    self.caption_d=d
    self.names_v = names
    self.names_d = names_d
    self.names_category_list=c

  def create_args(self,args,panel,constraints=None):
    if isinstance(args,arguments_set):
      self.test_consistency(args)
      return args
    args_v=self.conv_to_vector(args)
    if not constraints is None:
      constraints.within(args_v,True)	
      constraints.set_fixed(args_v)
    args_d=self.conv_to_dict(args_v)
    dict_string=[]
    for c in self.categories:
      s=[]
      captions=self.caption_d[c]
      a=args_d[c].flatten()
      for i in range(len(captions)):
        s.append(f"'{captions[i]}':{a[i]}")
      dict_string.append(f"'{c}':\n"+"{"+",\n".join(s)+"}")
    dict_string="{"+",\n".join(dict_string)+"}"
    return arguments_set(args_d, args_v, dict_string, self,panel)

  def test_consistency(self,args):
    #for debugging only
    m=self.positions_map
    for i in m:
      dict_arg=args.args_d[m[i][1]]
      if len(dict_arg.shape)==2:
        dict_arg=dict_arg[m[i][2]]
      if dict_arg[0]!=args.args_v[i]:
        raise RuntimeError("argument inconsistency")

  def get_name_ix(self,x,single_item=False):
    #returns name, list of indicies
    if x is None:
      return None, None
    if x in self.caption_v:
      if single_item:
        indicies=self.caption_v.index(x)
      else:
        indicies=[self.caption_v.index(x)]	
      return x,indicies
    elif x in self.positions and not single_item:
      indicies=list(self.positions[x])
      return x,indicies
    elif x in self.names_v:
      if single_item:
        indicies=self.names_v.index(x)
      else:
        indicies=[self.names_v.index(x)]	
      return x,indicies			
    try:
      name=self.caption_v[x]
    except Exception as e:
      raise RuntimeError(f"{e}. The identifier of an argument must be an integer or a string macthing a name in 'self.caption_v' or a category in 'self.positions'")
    if single_item:
      return name,x
    else:
      return name,[x]

  def set_init_regression(self, initargs,panel, default):
    usrargs =  panel.options.arguments.value
    if not usrargs is None:
      if type(usrargs)==str:
        try:
          usrargs = eval(usrargs.replace(" array"," np.array"))
        except NameError as e:
          if str(e)=="name 'array' is not defined":
            usrargs = eval(usrargs.replace("array"," np.array"))
      args = self.create_args(usrargs,panel)
      for c in args.args_d:
        initargs[c] = args.args_d[c]
      self.initial_user_defined = True
      return initargs['beta'], initargs['omega'][0,0]
    p, q, d, k, m=panel.pqdkm
    beta,rho,lmbda,u=ARMA_regression(panel)
    v=panel.var(u)
    if panel.options.EGARCH.value==0:
      initargs['omega'][0,0]=v
    else:
      initargs['omega'][0,0]=np.log(v)


    initargs['beta']=beta

    if default: #turns out calculating ARMA parameters serves no purpose, default is therefore allways true. These coefficients are easily found by linesearch for given regression coefficients. 
      if q > 0:
        initargs['lambda'][0] = lmbda

      if p > 0:
        initargs['rho'][0] = rho	
      
      if k > 0:
        initargs['gamma'][0] = 0	      

    if panel.options.fixed_random_variance_eff.value==0:
      if v<1e-20:
        print('Warning, your model may be over determined. Check that you do not have the dependent among the independents')	
    return beta, v

def set_GARCH(panel,initargs,u,m):
  matrices=logl.set_garch_arch(panel,initargs)
  if matrices is None:
    e=u
  else:
    AMA_1,AMA_1AR,GAR_1,GAR_1MA=matrices
    e = fu.dot(AMA_1AR,u)*panel.included[3]		
  h=h_func(e, panel,initargs)
  if m>0:
    initargs['gamma'][0]=0
    initargs['psi'][0]=0


def h_func(e,panel,initargs):
  z=None
  if len(initargs['z'])>0:
    z=initargs['z'][0]	
  h_val,h_e_val,h_2e_val,h_z,h_2z,h_e_z=logl.h(e,z,panel)
  return h_val*panel.included[3]



def ARMA_regression(panel):
  gfre=panel.options.fixed_random_group_eff.value
  tfre=panel.options.fixed_random_time_eff.value
  re_obj_i=re.re_obj(panel,True,panel.T_i,panel.T_i,gfre)
  re_obj_t=re.re_obj(panel,False,panel.date_count_mtrx,panel.date_count,tfre)
  X=(panel.X+re_obj_i.RE(panel.X, panel)+re_obj_t.RE(panel.X, panel))*panel.included[3]
  Y=(panel.Y+re_obj_i.RE(panel.Y, panel)+re_obj_t.RE(panel.Y, panel))*panel.included[3]
  beta,u=stat.OLS(panel,X,Y,return_e=True)
  rho,lmbda=ARMA_process_calc(u,panel)
  return beta,rho,lmbda,u

def ARMA_process_calc(e,panel):
  c=stat.correlogram(panel,e,2,center=True)[1:]

  rho = max(min(c[1]/c[0],0.99), -0.99)


  lmbda = 0
  try:
    lmbda1 = (1 - 2*c[0]*rho + rho**2)/(2*(c[0]-rho))
    lmbda2 = (( (rho**2 - 1)*(rho**2 - 1 + 4*c[0]**2 - 4*c[0]*rho) )**0.5) / (2*(c[0]-rho))

    if abs(lmbda1+lmbda2)>abs(lmbda1-lmbda2):
      lmbda = max(min(lmbda1 - lmbda2,0.99), -0.99)
    else:
      lmbda = max(min(lmbda1 + lmbda2,0.99), -0.99)
  except:
    pass


  return rho,lmbda


def add_names(T,captionstr,category,d,c,captions, names, names_d):
  a=[]
  n=[]
  if ' ' in captionstr:
    namestr = captionstr.split(' ')[0]
  for i in range(T):
    a.append(captionstr %(i,))
    n.append(namestr %(i,))
  captions.extend(a)
  names.extend(n)
  d[category]=a
  names_d[category]=n
  c.append(a)


class arguments_set:
  """A class that contains the numeric arguments used in the maximization
  in all shapes and forms needed."""
  def __init__(self,args_d,args_v,dict_string,arguments,panel):
    self.args_d=args_d#dictionary of arguments
    self.args_v=args_v#vector of arguments
    self.dict_string=dict_string#a string defining a dictionary of named arguments. For user input of initial arguments
    self.caption_v=arguments.caption_v#vector of captions
    self.caption_d=arguments.caption_d#dict of captions
    self.names_v=arguments.names_v#vector of names
    self.names_d=arguments.names_d#dict of names		
    self.n_args=len(self.args_v)
    self.pqdkm=panel.pqdkm
    self.positions=arguments.positions
    self.names_category_list=arguments.names_category_list
